/*
 *
 *  $Id: pvrusb2-debugifc.c,v 1.3 2005/06/19 22:27:13 isely Exp $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/string.h>
#include <linux/slab.h>
#include "pvrusb2-hdw-low.h"
#include "pvrusb2-debugifc.h"
#include "pvrusb2-debug.h"

struct debugifc_mask_item {
    const char *name;
    unsigned long msk;
};

static struct debugifc_mask_item mask_items[] = {
    {"TUNER_CFG_STD",PVR2_SUBSYS_TUNER_CFG_STD},
    {"TUNER_CFG_FREQ",PVR2_SUBSYS_TUNER_CFG_FREQ},
    {"AUDIO_CFG_VBBTM",PVR2_SUBSYS_AUDIO_CFG_VBBTM},
    {"AUDIO_CFG_STD",PVR2_SUBSYS_AUDIO_CFG_STD},
    {"AUDIO_CFG_MODE",PVR2_SUBSYS_AUDIO_CFG_MODE},
    {"DIGITIZER_CFG_NORM",PVR2_SUBSYS_DIGITIZER_CFG_NORM},
    {"DIGITIZER_CFG_INPUT",PVR2_SUBSYS_DIGITIZER_CFG_INPUT},
    {"DIGITIZER_CFG_SIZE",PVR2_SUBSYS_DIGITIZER_CFG_SIZE},
    {"DIGITIZER_CFG_AUDIO",PVR2_SUBSYS_DIGITIZER_CFG_AUDIO},
    {"DIGITIZER_CFG_BCSH",PVR2_SUBSYS_DIGITIZER_CFG_BCSH},
    {"ENC_CFG",PVR2_SUBSYS_ENC_CFG},
    {"DIGITIZER_RUN",PVR2_SUBSYS_DIGITIZER_RUN},
    {"USBSTREAM_RUN",PVR2_SUBSYS_USBSTREAM_RUN},
    {"ENC_RUN",PVR2_SUBSYS_ENC_RUN},
};


static unsigned int debugifc_count_whitespace(const char *buf,
                                              unsigned int count)
{
    unsigned int scnt;
    char ch;

    for (scnt = 0; scnt < count; scnt++) {
        ch = buf[scnt];
        if (ch == ' ') continue;
        if (ch == '\t') continue;
        if (ch == '\n') continue;
        break;
    }
    return scnt;
}


static unsigned int debugifc_count_nonwhitespace(const char *buf,
                                                 unsigned int count)
{
    unsigned int scnt;
    char ch;

    for (scnt = 0; scnt < count; scnt++) {
        ch = buf[scnt];
        if (ch == ' ') break;
        if (ch == '\t') break;
        if (ch == '\n') break;
    }
    return scnt;
}


static unsigned int debugifc_isolate_word(const char *buf,unsigned int count,
                                          const char **wstrPtr,
                                          unsigned int *wlenPtr)
{
    const char *wptr;
    unsigned int consume_cnt = 0;
    unsigned int wlen;
    unsigned int scnt;

    wptr = 0;
    wlen = 0;
    scnt = debugifc_count_whitespace(buf,count);
    consume_cnt += scnt; count -= scnt; buf += scnt;
    if (!count) goto done;

    scnt = debugifc_count_nonwhitespace(buf,count);
    if (!scnt) goto done;
    wptr = buf;
    wlen = scnt;
    consume_cnt += scnt; count -= scnt; buf += scnt;

 done:
    *wstrPtr = wptr;
    *wlenPtr = wlen;
    return consume_cnt;
}


static int debugifc_match_keyword(const char *buf,unsigned int count,
                                  const char *keyword)
{
    unsigned int kl;
    if (!keyword) return 0;
    kl = strlen(keyword);
    if (kl != count) return 0;
    return !memcmp(buf,keyword,kl);
}


static unsigned long debugifc_find_mask(const char *buf,unsigned int count)
{
    struct debugifc_mask_item *mip;
    unsigned int idx;
    for (idx = 0; idx < sizeof(mask_items)/sizeof(mask_items[0]); idx++) {
        mip = mask_items + idx;
        if (debugifc_match_keyword(buf,count,mip->name)) return mip->msk;
    }
    return 0;
}


static int debugifc_print_mask(char *buf,unsigned int sz,
                               unsigned long msk,unsigned long val)
{
    struct debugifc_mask_item *mip;
    unsigned int idx;
    int bcnt = 0;
    int ccnt;
    for (idx = 0; idx < sizeof(mask_items)/sizeof(mask_items[0]); idx++) {
        mip = mask_items + idx;
        if (!(mip->msk & msk)) continue;
        ccnt = scnprintf(buf,sz,"%s%c%s",
                         (bcnt ? " " : ""),
                         ((mip->msk & val) ? '+' : '-'),
                         mip->name);
        sz -= ccnt;
        buf += ccnt;
        bcnt += ccnt;
    }
    return bcnt;
}

static unsigned int debugifc_parse_subsys_mask(const char *buf,
                                               unsigned int count,
                                               unsigned long *mskPtr,
                                               unsigned long *valPtr)
{
    const char *wptr;
    unsigned int consume_cnt = 0;
    unsigned int scnt;
    unsigned int wlen;
    int mode;
    unsigned long m1,msk,val;

    msk = 0;
    val = 0;

    while (count) {
        scnt = debugifc_isolate_word(buf,count,&wptr,&wlen);
        if (!scnt) break;
        consume_cnt += scnt; count -= scnt; buf += scnt;
        if (!wptr) break;

        mode = 0;
        if (wlen) switch (wptr[0]) {
        case '+':
            wptr++;
            wlen--;
            break;
        case '-':
            mode = 1;
            wptr++;
            wlen--;
            break;
        }
        if (!wlen) continue;
        m1 = debugifc_find_mask(wptr,wlen);
        if (!m1) break;
        msk |= m1;
        if (!mode) val |= m1;
    }
    *mskPtr = msk;
    *valPtr = val;
    return consume_cnt;
}


int pvr2_debugifc_print_status(struct pvr2_hdw *hdw,
                               char *buf,unsigned int acnt)
{
    int bcnt = 0;
    int ccnt;
    unsigned long msk;
    int ret;

    ret = pvr2_hdw_is_hsm(hdw);
    ccnt = scnprintf(buf,acnt,"USB link speed: %s\n",
                     (ret < 0 ? "FAIL" : (ret ? "high" : "full")));
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;

    ccnt = scnprintf(buf,acnt,"Streaming is %s\n",
                     pvr2_hdw_get_streaming(hdw) ? "on" : "off");
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;

    msk = pvr2_hdw_subsys_get(hdw);
    ccnt = scnprintf(buf,acnt,"Subsystems enabled / configured: ");
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;
    ccnt = debugifc_print_mask(buf,acnt,msk,msk);
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;
    ccnt = scnprintf(buf,acnt,"\n");
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;
    ccnt = scnprintf(buf,acnt,"Subsystems disabled / unconfigured: ");
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;
    ccnt = debugifc_print_mask(buf,acnt,~msk,msk);
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;
    ccnt = scnprintf(buf,acnt,"\n");
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;

    msk = pvr2_hdw_subsys_stream_get(hdw);
    ccnt = scnprintf(buf,acnt,"Subsystems stopped on stream shutdown: ");
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;
    ccnt = debugifc_print_mask(buf,acnt,msk,msk);
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;
    ccnt = scnprintf(buf,acnt,"\n");
    bcnt += ccnt; acnt -= ccnt; buf += ccnt;

    return bcnt;
}


int pvr2_debugifc_do1cmd(struct pvr2_hdw *hdw,const char *buf,
                         unsigned int count)
{
    const char *wptr;
    unsigned int wlen;
    unsigned int scnt;

    scnt = debugifc_isolate_word(buf,count,&wptr,&wlen);
    if (!scnt) return 0;
    count -= scnt; buf += scnt;
    if (!wptr) return 0;

    pvr2_trace(PVR2_TRACE_DEBUGIFC,"debugifc cmd: \"%.*s\"",wlen,wptr);
    if (debugifc_match_keyword(wptr,wlen,"reset")) {
        scnt = debugifc_isolate_word(buf,count,&wptr,&wlen);
        if (!scnt) return -EINVAL;
        count -= scnt; buf += scnt;
        if (!wptr) return -EINVAL;
        if (debugifc_match_keyword(wptr,wlen,"cpu")) {
            pvr2_hdw_cpureset_assert(hdw,!0);
            pvr2_hdw_cpureset_assert(hdw,0);
            return 0;
        } else if (debugifc_match_keyword(wptr,wlen,"bus")) {
            pvr2_hdw_device_reset(hdw);
        } else if (debugifc_match_keyword(wptr,wlen,"soft")) {
            return pvr2_hdw_cmd_soft_reset(hdw);
        } else if (debugifc_match_keyword(wptr,wlen,"deep")) {
            return pvr2_hdw_cmd_deep_reset(hdw);
        }
        return -EINVAL;
    } else if (debugifc_match_keyword(wptr,wlen,"subsys_flags_chg")) {
        unsigned long msk = 0;
        unsigned long val = 0;
        if (debugifc_parse_subsys_mask(buf,count,&msk,&val) != count) {
            pvr2_trace(PVR2_TRACE_DEBUGIFC,
                       "debugifc parse error on subsys mask");
            return -EINVAL;
        }
        pvr2_hdw_subsys_bit_chg(hdw,msk,val);
        return 0;
    } else if (debugifc_match_keyword(wptr,wlen,"stream_flags_chg")) {
        unsigned long msk = 0;
        unsigned long val = 0;
        if (debugifc_parse_subsys_mask(buf,count,&msk,&val) != count) {
            pvr2_trace(PVR2_TRACE_DEBUGIFC,
                       "debugifc parse error on stream mask");
            return -EINVAL;
        }
        pvr2_hdw_subsys_stream_bit_chg(hdw,msk,val);
        return 0;
    }
    pvr2_trace(PVR2_TRACE_DEBUGIFC,
               "debugifc failed to recognize cmd: \"%.*s\"",wlen,wptr);
    return -EINVAL;
}


int pvr2_debugifc_docmd(struct pvr2_hdw *hdw,const char *buf,
                        unsigned int count)
{
    unsigned int bcnt = 0;
    int ret;

    while (count) {
        for (bcnt = 0; bcnt < count; bcnt++) {
            if (buf[bcnt] == '\n') break;
        }

        ret = pvr2_debugifc_do1cmd(hdw,buf,bcnt);
        if (ret < 0) return ret;
        if (bcnt < count) bcnt++;
        buf += bcnt;
        count -= bcnt;
    }

    return 0;
}


/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
